from .utils import *
from .skipgram import *
from .deepWalk import DeepWalk
from .node2vec import Node2vec
from .struc2vec import Struc2Vec
from .Classifiers import TrainingClassifiers

__version__ = "1.0.3"


from sklearn.neural_network import MLPClassifier
from sklearn.naive_bayes import MultinomialNB
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import GradientBoostingClassifier
from sklearn import preprocessing
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
from sklearn import tree
class TrainingClassifiers:
    def applyLogistic(self, X_train, y_train, X_test):
        clf_logreg = LogisticRegression()
        clf_logreg.fit(X_train, y_train)
        y_pred_class = clf_logreg.predict(X_test)
        print("LR completed")
        return y_pred_class

    def applyRandomForest(self, X_train, y_train, X_test):
        # clf_randomForest = RandomForestClassifier(n_estimators=382, criterion='entropy', max_features=116, max_depth=33, min_samples_split=5, min_samples_leaf=1 )
        clf_randomForest = RandomForestClassifier()
        clf_randomForest.fit(X_train, y_train)
        y_pred_class = clf_randomForest.predict(X_test)
        print("RF completed")
        return y_pred_class


    def apply_GradientBoosting(self, X_train, y_train, X_test):
        clf_gb = GradientBoostingClassifier()
        clf_gb.fit(X_train, y_train)
        y_pred_class = clf_gb.predict(X_test)
        print("GB completed")
        return y_pred_class

    def applyMLP(self, X_train, y_train, X_test):
        clf_MLP = MLPClassifier(solver='sgd')
        clf_MLP.fit(X_train, y_train)
        y_pred_class = clf_MLP.predict(X_test)
        print("MLP completed")
        return y_pred_class


    def applyDecisionTree(self, X_train, y_train, X_test):
        clf = tree.DecisionTreeClassifier()
        clf.fit(X_train, y_train)
        y_pred_class = clf.predict(X_test)
        print("DT completed")
        return y_pred_class

    def labelEnocder(self, labels):
        le = preprocessing.LabelEncoder()
        y = le.fit_transform(labels)
        return y

    def prepareTrainTestData(self, X, y, testSize):
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=testSize)
        return X_train, X_test, y_train, y_test

    def accuracy(self, y_true, y_pred):
        return accuracy_score(y_true, y_pred)


import warnings
warnings.filterwarnings("ignore")
import numpy as np
import scipy.sparse as sp
import ge
# Load Embedding
print("Load Embedding")
idsEmbeddingClsLabels = np.genfromtxt("../data/cora.embedding", dtype=np.dtype(str))

print("Classifier Training")
# Prepare data for training testing
labels = idsEmbeddingClsLabels[:, -1]
embedding = sp.csr_matrix(idsEmbeddingClsLabels[:, 1:-1], dtype=np.float32)


tr = ge.TrainingClassifiers()
y = tr.labelEnocder(labels)
# Prepare Train test data
X_train, X_test, y_train, y_test = tr.prepareTrainTestData(embedding, labels, 0.33)

# Choose one of the following classifier for training train a classifier

y_pred = tr.applyDecisionTree(X_train.toarray(), y_train, X_test.toarray())
y_pred = tr.applyLogistic(X_train.toarray(), y_train, X_test.toarray())
y_pred = tr.applyRandomForest(X_train.toarray(), y_train, X_test.toarray())
y_pred = tr.apply_GradientBoosting(X_train.toarray(), y_train, X_test.toarray())

y_pred = tr.applyMLP(X_train.toarray(), y_train, X_test.toarray())
print("Accuracy:", tr.accuracy(y_test, y_pred))



# import module
import ge

# Set Path to data
dataset = "cora"
data_dir = "../data"

# input graph
myGraph = ge.loadGraph(data_dir, dataset)

# Set Parameters
embedDim = 10 # embedding size
numbOfWalksPerVertex = 2 # walks per vertex
walkLength = 5 # walk lenght
lr =0.25 # learning rate
windowSize = 3 # window size

# # Set Parameters
# embedDim = 128 # embedding size
# numbOfWalksPerVertex = 10 # walks per vertex
# walkLength = 80 # walk lenght
# lr =0.025 # learning rate
# windowSize = 10 # window size




# DeepWalk
rw = ge.DeepWalk(myGraph, walkLength=walkLength, embedDim=embedDim, numbOfWalksPerVertex=numbOfWalksPerVertex, \
                 windowSize=windowSize, lr = lr)

# Skip Gram model
model_skip_gram = ge.SkipGramModel(rw.totalNodes, rw.embedDim)

print("Learning Embedding")
# Learning Node Embedding
model = rw.learnNodeEmbedding(model_skip_gram)

# Learning Edge Embedding
# model = dw.learnEdgeEmbedding(model_skip_gram)

# Plot Embedding
ge.plot_2DEmbedding(rw)

# Save embedding to disk
print("Saving embedding to disk")
ge.saveEmbedding(data_dir, dataset, rw)

print("Generating embedding for a node and edge")
node1 = 35
node2 = 40
# Get Embedding for a node
emb = rw.getNodeEmbedding(node1)
print("Node Embedding", emb.data)
#
# # Get Embedding for an edge
emb = rw.getEdgeEmbedding(node1, node2)
print("Edge Embedding", emb.data)






import pandas as pd
import numpy as np
import math
from fastdtw import fastdtw
import os
import networkx as nx
import matplotlib.pyplot as plt
import warnings
warnings.filterwarnings("ignore")
def chooseNeighbor(v, graphs, layers_alias, layers_accept, layer):

    v_list = graphs[layer][v]

    idx = alias_sample(layers_accept[layer][v], layers_alias[layer][v])
    v = v_list[idx]

    return v

def alias_sample(accept, alias):
    """
    :param accept:
    :param alias:
    :return: sample index
    """
    N = len(accept)
    i = int(np.random.random()*N)
    r = np.random.random()
    if r < accept[i]:
        return i
    else:
        return alias[i]

def preprocess_nxgraph(graph):
    node2idx = {}
    idx2node = []
    node_size = 0
    for node in graph.nodes():
        node2idx[node] = node_size
        idx2node.append(node)
        node_size += 1
    return idx2node, node2idx

def partition_dict(vertices, workers):
    batch_size = (len(vertices) - 1) // workers + 1
    part_list = []
    part = []
    count = 0
    for v1, nbs in vertices.items():
        part.append((v1, nbs))
        count += 1
        if count % batch_size == 0:
            part_list.append(part)
            part = []
    if len(part) > 0:
        part_list.append(part)
    return part_list

def create_alias_table(area_ratio):
    l = len(area_ratio)
    accept, alias = [0] * l, [0] * l
    small, large = [], []
    area_ratio_ = np.array(area_ratio) * l
    for i, prob in enumerate(area_ratio_):
        if prob < 1.0:
            small.append(i)
        else:
            large.append(i)

    while small and large:
        small_idx, large_idx = small.pop(), large.pop()
        accept[small_idx] = area_ratio_[small_idx]
        alias[small_idx] = large_idx
        area_ratio_[large_idx] = area_ratio_[large_idx] - \
                                 (1 - area_ratio_[small_idx])
        if area_ratio_[large_idx] < 1.0:
            small.append(large_idx)
        else:
            large.append(large_idx)

    while large:
        large_idx = large.pop()
        accept[large_idx] = 1
    while small:
        small_idx = small.pop()
        accept[small_idx] = 1

    return accept, alias


def cost(a, b):
    ep = 0.5
    m = max(a, b) + ep
    mi = min(a, b) + ep
    return ((m / mi) - 1)


def cost_min(a, b):
    ep = 0.5
    m = max(a[0], b[0]) + ep
    mi = min(a[0], b[0]) + ep
    return ((m / mi) - 1) * min(a[1], b[1])


def cost_max(a, b):
    ep = 0.5
    m = max(a[0], b[0]) + ep
    mi = min(a[0], b[0]) + ep
    return ((m / mi) - 1) * max(a[1], b[1])


def convert_dtw_struc_dist(distances, startLayer=1):
    for vertices, layers in distances.items():
        keys_layers = sorted(layers.keys())
        startLayer = min(len(keys_layers), startLayer)
        for layer in range(0, startLayer):
            keys_layers.pop(0)

        for layer in keys_layers:
            layers[layer] += layers[layer - 1]
    return distances


def get_vertices(v, degree_v, degrees, n_nodes):
    a_vertices_selected = 2 * math.log(n_nodes, 2)
    vertices = []
    try:
        c_v = 0

        for v2 in degrees[degree_v]['vertices']:
            if (v != v2):
                vertices.append(v2)  # same degree
                c_v += 1
                if (c_v > a_vertices_selected):
                    raise StopIteration

        if ('before' not in degrees[degree_v]):
            degree_b = -1
        else:
            degree_b = degrees[degree_v]['before']
        if ('after' not in degrees[degree_v]):
            degree_a = -1
        else:
            degree_a = degrees[degree_v]['after']
        if (degree_b == -1 and degree_a == -1):
            raise StopIteration  # not anymore v
        degree_now = verifyDegrees(degrees, degree_v, degree_a, degree_b)
        # nearest valid degree
        while True:
            for v2 in degrees[degree_now]['vertices']:
                if (v != v2):
                    vertices.append(v2)
                    c_v += 1
                    if (c_v > a_vertices_selected):
                        raise StopIteration

            if (degree_now == degree_b):
                if ('before' not in degrees[degree_b]):
                    degree_b = -1
                else:
                    degree_b = degrees[degree_b]['before']
            else:
                if ('after' not in degrees[degree_a]):
                    degree_a = -1
                else:
                    degree_a = degrees[degree_a]['after']

            if (degree_b == -1 and degree_a == -1):
                raise StopIteration

            degree_now = verifyDegrees(degrees, degree_v, degree_a, degree_b)

    except StopIteration:
        return list(vertices)

    return list(vertices)


def verifyDegrees(degrees, degree_v_root, degree_a, degree_b):

    if(degree_b == -1):
        degree_now = degree_a
    elif(degree_a == -1):
        degree_now = degree_b
    elif(abs(degree_b - degree_v_root) < abs(degree_a - degree_v_root)):
        degree_now = degree_b
    else:
        degree_now = degree_a

    return degree_now

def compute_dtw_dist(part_list, degreeList, dist_func):
    dtw_dist = {}
    for v1, nbs in part_list:
        lists_v1 = degreeList[v1]  # lists_v1 :orderd degree list of v1
        for v2 in nbs:
            lists_v2 = degreeList[v2]  # lists_v1 :orderd degree list of v2
            max_layer = min(len(lists_v1), len(lists_v2))  # valid layer
            dtw_dist[v1, v2] = {}
            for layer in range(0, max_layer):
                dist, path = fastdtw(
                    lists_v1[layer], lists_v2[layer], radius=1, dist=dist_func)
                dtw_dist[v1, v2][layer] = dist
    return dtw_dist

def operator_hadamard(u, v):
    return u * v

def plot_2DEmbedding(dw):
    xs = dw.model.W1.data[:, 0]
    ys = dw.model.W1.data[:, 1]
    ls = list(range(0, len(xs)))
    plt.scatter(xs, ys)
    for x,y,l in zip(xs,ys, ls):
        plt.annotate((dw.nodeEncoder.inverse_transform([l])[0]), (x, y))
    plt.show()

# Save embedding to disk
def saveEmbedding(data_dir, dataset, dw):
    idsFeaturesLabels = np.genfromtxt("{}{}.content".format(data_dir + "/", dataset), dtype=np.dtype(str))
    embCol = list()
    for i in range(dw.embedDim):
        embCol.append('c' + str(i))
    embCol.append('class_label')
    idsLabels = idsFeaturesLabels[:, [0, -1]]
    df = pd.DataFrame(columns = embCol,
                      index = list(idsLabels[:,0])) # adding nodes as an index

    for row in idsLabels:
        # Get embedding for a node
        f = dw.getNodeEmbedding(int(row[0])).tolist()
        f.append(row[1])
        df.loc[row[0]] = f
    df.to_csv("{}{}.embedding".format(data_dir + "/", dataset), sep='\t', header=False)

def loadGraph(data_dir, dataset):
    # Load Data
    print("Loading Data...")
    data_dir = os.path.expanduser(data_dir)
    edgelist = pd.read_csv(os.path.join(data_dir, dataset + ".cites"), sep='\t', header=None, names=["target", "source"])
    return nx.from_pandas_edgelist(edgelist)

# custom format warning with only a message
def custom_formatwarning(msg, *args, **kwargs):
    return str(msg) + '\n'
warnings.formatwarning = custom_formatwarning

